@@ -4,15 +4,34 @@ import json
 import logging
 from pydantic import BaseModel, Field, validate_arguments
 import requests
+from requests.adapters import HTTPAdapter
+from requests.packages.urllib3.util.retry import Retry
 import time
 from typing import Any, Dict, List, Optional
 from urllib.parse import quote_plus, urlparse
 
 
+class RotateSegment(BaseModel):
+    pages: str
+    degrees: int
+
+
+class Mutation(BaseModel):
+    rotate: Optional[int]
+    split: Optional[str]
+    remove_pages: Optional[str]
+    split_segments: Optional[List[str]]
+    rotate_segments: Optional[List[RotateSegment]]
+
+    def to_json(self):
+        return self.dict(exclude_unset=True)
+
+
 class FilePath(BaseModel):
     name: str
     path: str
     uid: Optional[str]
+    mutate: Optional[Mutation]
 
 
 class FieldType(str, Enum):
@@ -102,7 +121,14 @@ class Impira:
     def __init__(self, org_name, api_token, base_url="https://app.impira.com", ping=True):
         self.org_url = urljoin(base_url, "o", org_name)
         self.api_url = urljoin(self.org_url, "api/v2")
-        self.headers = {"X-Access-Token": api_token}
+        self.session = requests.Session()
+        self.session.headers.update({"X-Access-Token": api_token})
+
+        # Following a suggestion in https://stackoverflow.com/questions/23013220/max-retries-exceeded-with-url-in-requests
+        retry = Retry(connect=10, backoff_factor=0.5)
+        adapter = HTTPAdapter(max_retries=retry)
+        self.session.mount("http://", adapter)
+        self.session.mount("https://", adapter)
 
         if ping:
             try:
@@ -138,9 +164,8 @@ class Impira:
 
     @validate_arguments
     def update(self, collection_id: str, data: List[Dict[str, Any]]):
-        resp = requests.patch(
+        resp = self.session.patch(
             self._build_resource_url("fc", collection_id),
-            headers=self.headers,
             json={"data": data},
         )
 
@@ -151,9 +176,8 @@ class Impira:
 
     @validate_arguments
     def add_files_to_collection(self, collection_id: str, file_ids: List[str]):
-        resp = requests.post(
+        resp = self.session.post(
             self._build_resource_url("ec", "file_collection_contents"),
-            headers=self.headers,
             json={"data": [{"file_uid": u, "collection_uid": collection_id} for u in file_ids]},
         )
 
@@ -174,9 +198,8 @@ class Impira:
             )
 
         # Create collection is implemented as an empty insert
-        resp = requests.post(
+        resp = self.session.post(
             self._build_resource_url("collection", collection_name),
-            headers=self.headers,
             json={"data": []},
         )
 
@@ -192,9 +215,8 @@ class Impira:
 
     @validate_arguments
     def delete_field(self, collection_id: str, field_name: str):
-        resp = requests.delete(
+        resp = self.session.delete(
             urljoin(self.api_url, "schema/ecs/file_collections::%s/fields/%s" % (collection_id, field_name)),
-            headers=self.headers,
         )
 
         if not resp.ok:
@@ -202,9 +224,8 @@ class Impira:
 
     @validate_arguments
     def create_field(self, collection_id: str, field_spec: FieldSpec):
-        resp = requests.post(
+        resp = self.session.post(
             urljoin(self.api_url, "schema/ecs/file_collections::%s/fields" % (collection_id)),
-            headers=self.headers,
             json=dict(field_spec),
         )
 
@@ -213,9 +234,8 @@ class Impira:
 
     @validate_arguments
     def create_fields(self, collection_id: str, field_specs: List[FieldSpec]):
-        resp = requests.post(
+        resp = self.session.post(
             urljoin(self.api_url, "schema/ecs/file_collections::%s/fields" % (collection_id)),
-            headers=self.headers,
             json=[dict(field_spec) for field_spec in field_specs],
         )
 
@@ -239,7 +259,7 @@ class Impira:
         collection_id: str,
         from_collection_id: str,
     ):
-        resp = requests.post(
+        resp = self.session.post(
             urljoin(
                 self.api_url,
                 "schema",
@@ -248,7 +268,6 @@ class Impira:
                 "importfields",
                 "file_collections::" + from_collection_id,
             ),
-            headers=self.headers,
         )
 
         if not resp.ok:
@@ -277,7 +296,7 @@ class Impira:
         cursor = None
 
         must_see = set(uids)
-        while True:
+        while must_see:
             resp = self.query(query, mode="poll", cursor=cursor, timeout=60)
             for d in resp["data"] or []:
                 if d["action"] != "insert":
@@ -294,8 +313,33 @@ class Impira:
 
             cursor = resp["cursor"]
 
-            if len(must_see) == 0:
-                break
+    @validate_arguments
+    def fetch_split_results(self, upload_uids: List[str], timeout=60):
+        upload_uid_list = ", ".join(['"%s"' % u for u in upload_uids])
+        query = (
+            "@`files`[uid, upload_uid: File.upload_uid, child: -eq(uid, File.upload_uid)] in(File.upload_uid, %s) and child=true"
+            % (upload_uid_list)
+        )
+
+        cursor = None
+
+        must_see = set(upload_uids)
+        while must_see:
+            resp = self.query(query, mode="poll", cursor=cursor, timeout=timeout)
+            for d in resp["data"] or []:
+                if d["action"] != "insert":
+                    continue
+
+                uid = d["data"]["uid"]
+                upload_uid = d["data"]["upload_uid"]
+                child = d["data"]["child"]
+
+                # This is an upload uid that has been processed, so we can remove it from the must_see list
+                if upload_uid in must_see:
+                    must_see.remove(upload_uid)
+                yield uid
+
+            cursor = resp["cursor"]
 
     @validate_arguments
     def query(self, query: str, mode: str = "iql", cursor: str = None, timeout: int = None):
@@ -306,9 +350,8 @@ class Impira:
         if timeout is not None:
             args["timeout"] = timeout
 
-        resp = requests.post(
+        resp = self.session.post(
             urljoin(self.api_url, mode),
-            headers=self.headers,
             json=args,
         )
         if not resp.ok:
@@ -331,10 +374,10 @@ class Impira:
                     raise InvalidRequest("Unsupported: specifying a UID in a multi-part file upload (%s)" % (f.uid))
 
             files_body = [("file", open(f.path, "rb")) for f in files]
-            data_body = [("data", json.dumps(_build_file_object(f.name, None, f.uid))) for f in files]
-            resp = requests.post(
+            data_body = [("data", json.dumps(_build_file_object(f.name, None, f.uid, f.mutate))) for f in files]
+
+            resp = self.session.post(
                 self._build_collection_url(collection_id, use_async=True),
-                headers=self.headers,
                 files=tuple(files_body),
                 data=tuple(data_body),
             )
@@ -344,26 +387,34 @@ class Impira:
             elif not resp.ok:
                 raise APIError(resp)
             else:
-                return resp.json()["uids"]
+                return self._handle_upload_response(resp)
 
     @validate_arguments
     def _upload_url(self, collection_id: Optional[str], files: List[FilePath]):
-        resp = requests.post(
+        resp = self.session.post(
             self._build_collection_url(collection_id, use_async=True),
-            headers=self.headers,
-            json={"data": [_build_file_object(f.name, f.path, f.uid) for f in files]},
+            json={"data": [_build_file_object(f.name, f.path, f.uid, f.mutate) for f in files]},
         )
         if not resp.ok:
             raise APIError(resp)
 
-        return resp.json()["uids"]
+        return self._handle_upload_response(resp)
+
+    def _handle_upload_response(self, upload_response: requests.Response):
+        resp_json = upload_response.json()
+
+        if resp_json.get("uids", None) is not None:
+            return resp_json.get("uids")
+        elif resp_json.get("upload_uids", None) is not None:
+            return self.fetch_split_results(resp_json.get("upload_uids"))
+
+        return []
 
     @validate_arguments
     def _set_field_path(self, entity_class: str, uid: str, path: List[str], data):
-        resp = requests.post(
+        resp = self.session.post(
             urljoin(self.api_url, "data", entity_class, uid, *[quote_plus(segment) for segment in path]),
-            headers=self.headers,
-            json={"data": data}
+            json={"data": data},
         )
 
         if not resp.ok:
@@ -400,12 +451,14 @@ class Impira:
         return base_url
 
 
-def _build_file_object(name, path=None, uid=None):
+def _build_file_object(name, path=None, uid=None, mutate=None):
     ret = {"File": {"name": name}}
     if path is not None:
         ret["File"]["path"] = path
     if uid is not None:
         ret["uid"] = uid
+    if mutate is not None:
+        ret["File"]["mutate"] = mutate.to_json()
     return ret
 
 
