@@ -8,7 +8,7 @@ from typing import Any, Dict, List, Optional, Tuple, Type, Union
 import time
 from uuid import uuid4
 
-from impira import Impira as ImpiraAPI, FieldType, InferredFieldType, parse_date
+from impira import Impira as ImpiraAPI, FieldType, InferredFieldType, parse_date, APIError
 
 from ..cmd.utils import environ_or_required
 from ..schema import schema_to_model
@@ -295,7 +295,7 @@ def data_projection(skip_downloading_text):
 def upload_and_retrieve_text(conn, collection_uid, f, skip_downloading_text):
     existing = conn.query(
         "@`file_collections::%s`%s name='%s' File.IsPreprocessed=true"
-        % (collection_uid, data_projection(skip_downloading_text), f["name"]),
+        % (collection_uid, data_projection(skip_downloading_text), f["name"].replace("'", "\\'")),
     )["data"]
     if len(existing) > 0:
         return existing[0]
@@ -349,7 +349,6 @@ def fields_to_doc_schema(fields) -> DocSchema:
             path = ["Label", "Value"]
             if trainer in (InferredFieldType.checkbox, InferredFieldType.signature):
                 path.append("Value")
-                
 
             scalar_type = find_path(f, *path)["fieldType"]
             if trainer == InferredFieldType.document_tag:
@@ -408,11 +407,18 @@ def row_to_record(row, doc_schema: DocSchema) -> Any:
 
 @validate_arguments
 def row_to_fname(row, use_original_filename) -> str:
+    file_name = row["File"].get("name") or ""
+    uid = row["uid"]
+
     if use_original_filename:
-        return row["File"]["name"]
+        return file_name
     else:
-        fname, ext = row["File"]["name"].rsplit(".", 1)
-        return fname + "-" + row["uid"] + "." + ext
+        if "." not in file_name:
+            fname = file_name + "-" + uid
+        else:
+            fname_, ext = file_name.rsplit(".", 1)
+            fname = fname_ + "-" + uid + "." + ext
+        return fname.replace("/", "_")
 
 
 class Impira(Tool):
@@ -495,7 +501,7 @@ class Impira(Tool):
                 file_data = [
                     x
                     for x in t.map(
-                        lambda f: upload_and_retrieve_text(conn, collection_uid, f, skip_downloading_text),
+                        lambda f: upload_and_retrieve_text(self._conn(), collection_uid, f, skip_downloading_text),
                         files,
                     )
                 ]
@@ -678,20 +684,32 @@ class Impira(Tool):
         )
 
         log.info("Running update on %d files" % len(labeled_files))
-        conn.update(
-            collection_uid,
-            [
-                {
-                    **{"uid": fd["uid"]},
-                    **{
-                        field_path: label.dict(exclude_none=True)
-                        for field_path, label in ld.items()
-                        if field_path in field_names_to_update
-                    },
-                }
-                for (fd, ld) in zip(labeled_files, labels)
-            ],
-        )
+
+        # Batch the updates into chunks of 200, and retry a few times because of deadlock-issues
+        for b in batch([x for x in zip(labeled_files, labels)], n=200):
+            for i in range(10):
+                if i > 0:
+                    log.warning("Sleeping for 1 second...")
+                    time.sleep(1)
+                try:
+                    conn.update(
+                        collection_uid,
+                        [
+                            {
+                                **{"uid": fd["uid"]},
+                                **{
+                                    field_path: label.dict(exclude_none=True)
+                                    for field_path, label in ld.items()
+                                    if field_path in field_names_to_update
+                                },
+                            }
+                            for (fd, ld) in b
+                        ],
+                    )
+                    break
+                except APIError as e:
+                    log.warning("Failed to update. Will retry up to %d more times" % (10 - i - 1))
+
         log.info("Done running update on %d files. Models will now update!" % len(labeled_files))
 
         # This code is currently too brittle to be useful, since model versions aren't a reliable way of knowing
@@ -741,11 +759,11 @@ class Impira(Tool):
         return doc_schema, records
 
     @validate_arguments
-    def snapshot_collections(self, use_original_filenames=False):
+    def snapshot_collections(self, use_original_filenames=False, max_files_per_collection=-1, num_samples=2):
         log = self._log()
 
         conn = self._conn()
-        files = conn.query("@files[uid, File: File[download_url, name]] -`File type`=Data")["data"]
+        files = conn.query("@files[uid, File: File[download_url, name]] -File.download_url=null -`File type`=Data")["data"]
         collections = conn.query(
             "@file_collection_contents[collection_uid, files: array_agg(file_uid)] -collection=null"
         )["data"]
@@ -758,7 +776,8 @@ class Impira(Tool):
 
         file_membership = {}
         for c in collections:
-            for f in c["files"]:
+            collection = c["files"] if max_files_per_collection < 0 else c["files"][:max_files_per_collection]
+            for f in collection:
                 if f not in file_membership:
                     file_membership[f] = []
                 file_membership[f].append(c["collection_uid"])
@@ -766,7 +785,16 @@ class Impira(Tool):
         sampled = {}
         # For each collection, pick up to two files that belong to that collection
         for c in collections:
-            for f in [f for f in c["files"] if len(file_membership[f]) == 1][:2]:
+            valid_files = [f for f in c["files"] if f in file_membership and len(file_membership[f]) == 1]
+            if num_samples > len(valid_files):
+                log.warning(
+                    f"The collection {c['collection_uid']} has fewer files than the requested number of samples "
+                    f"({len(valid_files)} < {num_samples}), so sampling all files in the collection. It's "
+                    "recommended that you either lower the number of samples or add more files to the collection."
+                )
+
+            num_samples_ = min(len(valid_files), num_samples)
+            for f in valid_files[:num_samples_]:
                 sampled[f] = c["collection_uid"]
 
         doc_schema = DocSchema(
@@ -783,7 +811,8 @@ class Impira(Tool):
                 "record": {
                     "Doc tag": DocumentTagLabel(value=collection_names[file_membership[row["uid"]][0]])
                     if row["uid"] in file_membership and len(file_membership[row["uid"]]) == 1
-                    else DocumentTagLabel(value=None) if row["uid"] not in file_membership 
+                    else DocumentTagLabel(value=None)
+                    if row["uid"] not in file_membership
                     else None,
                     "Sampled tag": DocumentTagLabel(value=collection_names[sampled[row["uid"]]])
                     if row["uid"] in sampled
@@ -791,6 +820,7 @@ class Impira(Tool):
                 },
             }
             for row in files
+            if row["uid"] in file_membership
         ]
 
         assert len(records) == len(set([r["name"] for r in records])), "Expected each filename to be unique"
